/*
 * Created on 24 fvr. 2005
 *
 */
package util.saveObject;

import java.io.*;
import java.util.HashMap;

/**
 * <p>Cette classe permet la sauvegarde d'objet au format XML, et constitue donc une 
 * alternative a la srialisation. Pour qu'un objet soit conserv, il doit implmenter
 * l'interface "Savable" et possder un constructeur par dfaut (pour la reconstruction
 * de l'objet).
 * 
 * <p>C'est la classe "InputSaveObject" qui se charge de reconstruire et restituer les 
 * objets dans l'tat o ils avaient t "sauvegards".
 * <p>Utilisation : 
 * Elle se dcompose en 3 points : cration d'un objet OuputSaveObject, ajout des
 * objets  conserver (add), et sauvegarde au format XML (saveInto).
 * 
 * <p>Il est aussi important de prciser que l'ordre d'ajout des objets sera l'ordre
 * de rcupration des objets (mthode getNext de InputSaveObject)
 * 
 * <p>Exemple : <br>
 * <pre>
 * class Class1 extends Savable {
 * 		String str;
 * 		int    nb;
 * 
 * 		public c1 () {
 * 			str = "texte";
 * 			nb = 5;
 * 		}
 * 
 * 		public void saveInto (OutputSaveObject sObj) throws SaveObjectException
 * 		{
 * 			sObj.add ("str", str);
 * 			sObj.add ("nb",  new Integer(nb));
 * 		}
 * 
 * 		public void loadFrom (InputSaveObject sObj) throws SaveObjectException
 * 		{
 * 			str = (String) sObj.getNext ("texte");
 * 			str = ((Integer) sObj.getNext ("nb")).intValue ();
 * 		}
 * }
 * </pre>
 * <p>Une telle classe pourra alors tre sauvegarde de la manire suivante :
 * <pre>
 * Class1 c1 = new Class1 (); 
 * OutputSaveObject sObj = new OutputSaveObject ();
 * FileOutputStream fos = new FileOutputStream ("fichier.xml");
 * sObj.add ("c1", c1);
 * sObj.saveInto (fos);
 * </pre>
 * 
 * @author Benjamin LE BOZEC
 */
public class OutputSaveObject extends SaveObject {


	/*
	 * refTable pointe vers un objet commun  tous les SaveObject impliqus
	 * dans une mme sauvegarde. Il permet d'associ a chaque objet sauvegard
	 * une rfrence unique (sous forme de chaine de caractre).
	 */
	private RefTable  refTable;

	/**
	 * Cration d'une nouvelle instance de "OutputSaveObject"
	 */
	public OutputSaveObject () {
		this (new RefTable());
		isRacine = true;
	}
	
	/**
	 * Ce constructeur est utilis par la classe elle-meme.
	 * @param refTable la table des rferences a utiliser par l'objet
	 */
	protected OutputSaveObject (RefTable refTable) {
		super ();
		this.refTable = refTable;
	}

	/**
	 * Permet d'ajouter un objet  sauvegarder.
	 * 
	 * @param name Nom de l'objet. Il ne doit pas forcment tre unique et ne consistue
	 * pas une rfrence. Il permet juste une meilleure clart dans le fichier XML.
	 * @param obj Objet a sauvegarder. Si ce n'est pas un objet primitif (Integer,String,..),
	 * il doit implmenter l'interface Savable
	 * 
	 * @throws SaveObjectException Si l'objet n'est ni de type primitif ni de type Savable 
	 */
	public void add (String name, Object obj) throws SaveObjectException {
		ElemSaveObject item = new ElemSaveObject ();
		item.name = name;
		item.className = obj.getClass().getName();
		
		if (isPrimitive (obj.getClass())) {
			// si l'objet est de type primitif (int, char, string,..), on 
			// l'enregistre tel quel, sans cration de rfrence
			item.value = obj.toString ();
			item.reference = null;
		} else if  (obj instanceof Savable) {
			// si l'objet implmente Savable, on rcupere le SaveObject qui lui
			// est associ
			
			if (refTable.containsObject(obj)) {
				// si l'objet est dja rfrenc, on rcupre juste sa rfrence
				item.value = null;
				item.reference = refTable.getRef(obj);
			} else {
				// sinon, on rcupere le SaveObject,
				// et on cr une rfrence
				item.reference = refTable.createRef(obj);
				OutputSaveObject sObj = new OutputSaveObject (refTable);
				((Savable) obj).saveInto(sObj);
				item.value = sObj;
			}
		} else 
			throw new SaveObjectException ("L'objet " + obj + " ne peut etre sauvegarder : il n'est n'y de type primitif, ni de type Savable");
		
		data.addLast(item);
	}
	
	/**
	 * Cette mthode permet de stocker un tableau en utilisant la class SavableArray
	 * @param name nom du tableau
	 * @param tab
	 * @throws SaveObjectException cf mthode 'add'
	 */
	public void add (String name, Object[] tab) throws SaveObjectException {
		SavableArray sTab = new SavableArray (tab);
		add (name, sTab);
	}

	/**
	 * Cette mthode enregistre tous les objets ajouts par la mthode "add" dans un
	 * fichier au format XML.
	 * 
	 * @param out flux dans lequel doivent tre crites les dones.
	 * @throws IOException
	 */
	public void saveInto (OutputStream out) throws IOException {
		XmlWriter writer = new XmlWriter (out);
		XmlPropertyList prop = new XmlPropertyList ();
		
		if (isRacine) {
			// On crit dans le fichier l'entete Xml ainsi qu'une balise <java> 
			writer.writeHeader();
			writer.writeBalise("java", false, true);
		}
		
		int i;
		ElemSaveObject item;
		
		for (i=0;i<data.size();i++) {
			// Rcupration et criture de tous les lments que le SaveObject contient
			item = (ElemSaveObject) data.get(i);
			prop.flush ();
			if (item.name!=null) prop.add ("name", item.name);
			if (item.reference!=null) prop.add ("ref", item.reference);
			if (item.value==null) {
				// Si cet item ne pointe vers aucun objet, c'est qu'il consistue
				// une simple rfrence vers un objet dja crits dans le fichier
				prop.add("class", item.className);
				writer.writeBalise("Object", prop, true, true);
			} else {
				if (item.value instanceof SaveObject) {
					prop.add("class", item.className);
					// Si l'objet est un SaveObject, il faut alors crire dans
					// le fichier les donnes qu'il contient 
					writer.writeBalise ("Object", prop, false, true);
					((OutputSaveObject) item.value).saveInto(out);
					writer.closeBalise ();
				} else {
					// sinon, il s'agit d'un objet "primitif"
					String defaultName = "java.lang";
					String littleClassName = item.className.substring(defaultName.length()+1);
					writer.writeBalise (littleClassName, prop, false, false);
					writer.writeContent ((String) item.value);
					writer.closeBalise ();
				}
			}
		}
		
		if (isRacine) {
			writer.closeBalise();
		}
	}
}

/**
 * Cette classe permet la cration et la gestion de rfrence  un objet.
 * On entend par rfrence une chaine de caractre unique pour un ensemble
 * d'objet donn. Cette rfrence est cre par concatnation entre le nom 
 * de la classe de l'objet ainsi qu'un numro.
 * 
 * Exemple : pour un objet de classe DataSync, une rference possible est "DataSync1".
 * 
 * @author Benjamin LE BOZEC
 *
 */
class RefTable {
	
	/*
	 * Table des rfrences :  un objet est associ une rference
	 */
	private HashMap refTable;
	
	/*
	 * Table des classes :  une classe est associe un compteur 
	 */
	private HashMap classTable; 
	
	/**
	 * Constructeur - initialisation des attributs
	 */
	protected RefTable () {
		refTable = new HashMap (10);
		classTable = new HashMap (10);
	}
	
	/**
	 * Teste la prsence de l'objet dans la table des rfrences
	 * @param obj
	 * @return vrai si l'objet y est, faux sinon
	 */
	public boolean containsObject (Object obj) {
		return refTable.containsKey(obj);
	}
	
	/**
	 * Permet de rcuprer la rfrence de l'objet key.
	 * @param key
	 * @return La rfrence si l'objet existe, null sinon
	 */
	public String getRef (Object key) {
		return (String) refTable.get(key);
	}
	
	/**
	 * Cr une nouvelle rference unique en fonction de l'objet donn en paramtre.
	 * La rfrence cre est une concatnation entre une chaine de caractre 
	 * (correspondant au nom de la classe  laquelle appartient l'objet) et un numro.
	 * Afin d'assurer l'unicit de cette rfrence, un compteur spcifique  chaque
	 * classe est utilis.
	 * 
	 * @param key L'objet dont une rfrence doit etre cre
	 * @return la rfrence cre
	 */
	public String createRef (Object key) {
		
		// rcupration du compteur de la classe (s'il existe)
		Integer cpt = (Integer) classTable.get(key.getClass());
		if (cpt==null)
			cpt = new Integer (1);
		else
			cpt = new Integer (cpt.intValue()+1);
		
		// mise  jour de la tables des classes
		classTable.put(key.getClass(), cpt);
		
		// cration de la rference
		String ref =key.getClass().getName() + cpt;
		
		// mise  jour de la table des rfrences
		refTable.put(key, ref);
		
		return ref;
	}
	
}

